/******************************************************************************
 *
 *       PIC18 C18 Secure Digital and Multimedia Cards Interface
 *
 ******************************************************************************
 * FileName:        FAT15.c
 * Dependencies:    GenericTypeDefs.h
 *					FAT16.h
 *					sdmmc.h
 *					string.h
 *					stdlib.h
 *					ctype.h
 *					_FATDefs.h
 * Processor:       PIC18/PIC24
 * Compiler:        C18/C30
 * Company:         Microchip Technology, Inc.
 *
 * Software License Agreement
 *
 * The software supplied herewith by Microchip Technology Incorporated
 * (the CompanyE for its PICmicro Microcontroller is intended and
 * supplied to you, the Companys customer, for use solely and
 * exclusively on Microchip PICmicro Microcontroller products. The
 * software is owned by the Company and/or its supplier, and is
 * protected under applicable copyright laws. All rights are reserved.
 * Any use in violation of the foregoing restrictions may subject the
 * user to criminal sanctions under applicable laws, as well as to
 * civil liability for the breach of the terms and conditions of this
 * license.
 *
 * THIS SOFTWARE IS PROVIDED IN AN AS ISECONDITION. NO WARRANTIES,
 * WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
 * TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
 * IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
 * CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 *
 *****************************************************************************/

#include "FAT16.h"
#ifdef USE_COMPACTFLASH
	#include "CF.h"
#elif defined USE_SDMMC
	#include "sdmmc.h"
	#include "sdmmc_lld.h"
#endif
#include "GenericTypeDefs.h"
#include "string.h"
#include "stdlib.h"
#include "ctype.h"
#include "_FATDefs.h"

#ifdef FAT16_DYNAMIC_MEM
	#ifdef USE_PIC18
		#include "salloc.h"
	#endif
#endif

// Define arrays for file allocation
#ifndef FAT16_DYNAMIC_MEM
	FILE	gFileArray[FAT16_MAX_FILES_OPEN];
	BYTE 	gFileSlotOpen[FAT16_MAX_FILES_OPEN];
#endif

DWORD lastFATSectorRead = 0xFFFF;
BYTE needWrite = FALSE;

// DOS directory entry structure
typedef struct __DIRENTRY
{
        char        DIR_Name[DIR_NAMESIZE];      // Name
        char        DIR_Extension[DIR_EXTENSION]; // Extension
        BYTE          DIR_Attr;         // Attributes
        BYTE          DIR_NTRes;        // reserved by NT
        BYTE          DIR_CrtTimeTenth; // millisecond stamp
        WORD         DIR_CrtTime;      // time created
        WORD         DIR_CrtDate;      // date created
        WORD         DIR_LstAccDate;   // Last Access date
        WORD         DIR_FstClusHI;    // high word of this enty's first cluster number
        WORD         DIR_WrtTime;      // last update time
        WORD         DIR_WrtDate;      // last update date
        WORD         DIR_FstClusLO;    // low word of this entry's first cluster number
        DWORD         DIR_FileSize;     // file size in DWORD
}_DIRENTRY;

typedef _DIRENTRY * DIRENTRY;

// directory entry management
#define DIR_ESIZE   32      // size of a directory entry (BYTEs)
#define DIR_NAME    0       // offset of file name
#define DIR_EXT     8       // offset of file extension
#define DIR_ATTRIB  11      // offset of attribute( 00ARSHDV) (BYTE)
#define DIR_TIME    22      // offset of last use time  (WORD)
#define DIR_DATE    24      // offset of last use date  (WORD)
#define DIR_CLST    26      // offset of first cluster in FAT (WORD)
#define DIR_SIZE    28      // offset of file size (DWORD)
#define DIR_DEL     0xE5    // marker of a deleted entry
#define DIR_EMPTY   0       // marker of last entry in directory

// number of directory entries in one sector
#define DIRENTRIES_PER_SECTOR   0x10

#define RAMwrite( a, f, d) *(a+f) = d
#define RAMread( a, f)  *(a+f)
#define RAMreadW( a, f) *(WORD *)(a+f)
#define RAMreadD( a, f) *(DWORD *)(a+f)

FILE *gBufferOwner = NULL;
DWORD lastClusterRead = 0xFFFFFFFF;

// internal errors
#define CE_FAT_EOF              60   // fat attempt to read beyond EOF
#define CE_EOF                  61   // reached the end of file   

// since we use an address generator, FILE is not actually the cast of what we pass
typedef FILE        * FILEOBJ;

#define FILEEXTPAGESIZE         sizeof(_FILEEXTND)

#define SIZEOF(s,m)             ((size_t) sizeof(((s *)0)->m)) 


// Private functions

WORD FATread (DISK *dsk, WORD ccls);
DIRENTRY Cache_File_Entry( FILEOBJ fo, WORD * curEntry, BYTE ForceRead);
BYTE Fill_File_Object(FILEOBJ fo, WORD *fHandle);
DWORD Cluster2Sector(DISK * disk, WORD cluster);
BYTE FAT_erase_cluster_chain (WORD cluster, DISK * dsk);
WORD FATfindEmptyCluster(FILEOBJ fo);
DIRENTRY LoadDirAttrib(FILEOBJ fo, WORD *fHandle);
void IncrementTimeStamp(DIRENTRY dir);
BYTE FindEmptyEntries(FILEOBJ fo, WORD *fHandle);
BYTE PopulateEntries(FILEOBJ fo, char *name , WORD *fHandle);
CETYPE FILECreateHeadCluster( FILEOBJ fo, WORD *cluster);
BYTE EraseCluster(DISK *disk, WORD cluster);
void FileObjectCopy(FILEOBJ foDest,FILEOBJ foSource);
CETYPE CreateFirstCluster(FILEOBJ fo);
BYTE isShort ( char aChar);
extern BYTE WriteProtectState(void);
BYTE ValidateChars (char * FileName);
BYTE FormatFileName( const char* fileName, char* fN2 );
WORD FATwrite (DISK *dsk, WORD ccls, WORD value, BYTE forceWrite);

CETYPE __fread( FILE *fo, BYTE * dest, WORD count, WORD *pBytesRead);
CETYPE __fclose(FILEOBJ fo);
CETYPE __fwrite( FILE* fo, void * src, WORD count, WORD* pBytesWritten );
CETYPE __fseek (FILEOBJ fo, DWORD offset, BYTE origin);
FILE * __fopen (const char * fileName, const char *mode, FILE* filePtr );

CETYPE CreateFileEntry(FILEOBJ fo, WORD *fHandle);
CETYPE FILEfind( FILEOBJ foDest, FILEOBJ foCompareTo, BYTE cmd);
BYTE FILEget_next_cluster(FILEOBJ fo, WORD n);
BYTE FILEallocate_new_cluster( FILEOBJ fo);
CETYPE FILEopen (FILEOBJ fo, WORD *fHandle, char type);
CETYPE FILEerase( FILEOBJ fo, WORD *fHandle, BYTE EraseClusters);

extern void Delayms(BYTE milliseconds);

/******************************************************************************
 * Function:        int FAT16Init()
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          int:	True if initialization successful, otherwise false
 *
 * Side Effects:    None
 *
 * Overview:        Initialize the static memory slots for holding
 *					file structures; only used when FAT16_DYNAMIC_MEM is
 *					not defined
 *
 * Note:            None
 *****************************************************************************/
int FAT16Init()
{
#ifndef FAT16_DYNAMIC_MEM
	int fIndex;

	for( fIndex = 0; fIndex < FAT16_MAX_FILES_OPEN; fIndex++ )
		gFileSlotOpen[fIndex] = TRUE;

#else
	#ifdef USE_PIC18
		SRAMInitHeap();
	#endif
#endif

	return MediaInit();
}

/******************************************************************************
 * Function:        FILE * fopen (const char * fileName, const char *  mode)
 *
 * PreCondition:    For read or append mode, file exists
 *
 * Input:           fileName	- The name of the file to open (stored in RAM)
 *					mode	- WRITE - Create a new file or replace an existing file
 *							  READ - Read data from an existing file
 *							  APPEND - Append data to an existing file
 *
 * Output:			FILE *	- The pointer to the file object
 *
 * Side Effects:    None
 *
 * Overview:        User called file open function (with fileName and mode 
 *					defined in RAM)
 *
 * Note:            None
 *****************************************************************************/
FILE * fopen( const char * fileName, const char *mode )
{

	FILE* 	filePtr;
	FILE* 	pFile;
	int		fIndex;

#ifdef FAT16_DYNAMIC_MEM
	filePtr = (FILE *) FAT16_malloc(sizeof(FILE));
#else

	filePtr = NULL;

	//Pick available file structure
	for( fIndex = 0; fIndex < FAT16_MAX_FILES_OPEN; fIndex++ )
	{
		if( gFileSlotOpen[fIndex] )	//this slot is available
		{
			gFileSlotOpen[fIndex] = FALSE;
			filePtr = &gFileArray[fIndex];
			break;
		}
	}

	if( filePtr == NULL )
		return NULL;		//no file structure slot available

#endif

	//try to open the file
	pFile = __fopen( fileName, mode, filePtr );

#ifdef FAT16_DYNAMIC_MEM

	if( pFile == NULL )
		FAT16_free( (unsigned char *)filePtr );
#else

	if( pFile == NULL )
		gFileSlotOpen[fIndex] = TRUE;	//put this slot back to the pool

#endif

	return pFile;
}

/******************************************************************************
 * Function:        CETYPE FILEfind( FILEOBJ foDest, FILEOBJ foCompareTo, BYTE cmd)
 *
 * PreCondition:    DISKmount function has been executed, foCompareTo loaded with
 *					file name
 *
 * Input:           foDest			- FILEOBJ containing information of file found
 *					foCompareTo		- FILEOBJ containing name of file to be found
 *					cmd				- 1 - search for a matching entry
 *									  2 - search for an empty entry
 *                  
 * Output:          CE_GOOD				- File found
 *                  CE_FILE_NOT_FOUND	- File not found 
 *
 * Side Effects:    None
 *
 * Overview:        Find a file given a name as passed via foCompareTo, place found in foDest
 *
 * Note:            None
 *****************************************************************************/

CETYPE FILEfind( FILEOBJ foDest, FILEOBJ foCompareTo, BYTE cmd)
{
    WORD	attrib,fHandle = 0;            // current entry counter
    BYTE	state,index;              // state of the current object
    CETYPE	statusB = CE_FILE_NOT_FOUND;           
    BYTE	character,test;

    // reset the cluster
    foDest->dirccls = foDest->dirclus;

    if(	Cache_File_Entry(foDest, &fHandle, TRUE) == NULL)
    {
        statusB = CE_BADCACHEREAD;
    }
    else
    {
        // Loop until you reach the end or find the file 
		while(1)
		{
			if(statusB!=CE_GOOD)
			{
				state = Fill_File_Object(foDest, &fHandle);
				if(state == NO_MORE)
				{
					break;
				}
			}
			else
			{
				break;
			}

            if(state == FOUND)
            {
                /* We got something */
                // get the attributes
                attrib = foDest->attributes;
            
           		attrib &= ATTR_MASK;
        
                // see if we are a volume id or hidden, ignore
                if((attrib != ATTR_VOLUME) && (attrib & ATTR_HIDDEN) != ATTR_HIDDEN)
                {
                    statusB = CE_GOOD;
                    character = (BYTE)'m'; // random value
                    
                    // search for one. if status = TRUE we found one
                    for(index = 0; (statusB == CE_GOOD) && index < DIR_NAMECOMP; index++)
                    {
                        // get the source character 
                        character = foDest->name[index];
                        // get the destination character
                        test = foCompareTo->name[index];
                        if(tolower(character) != tolower(test))
                            statusB = CE_FILE_NOT_FOUND; // Nope its not a match
                            
                     }// for loop
                     
               } // not dir nor vol
                    
            } // not found
            else 
            {
                /*** looking for an empty/re-usable entry ***/
                if ( cmd == 0) 
                    statusB = CE_GOOD;
            } // found or not    
            
            // just increment it no matter what happened
            fHandle++;
              
        }// while 
    }
    
    return(statusB);
} // FILEFind




/******************************************************************************
 * Function:        CETYPE FILEopen (FILEOBJ fo, WORD *fHandle, char type)
 *
 * PreCondition:    FILEfind returns true for read or append, fo contains file data
 *
 * Input:           fo			- File to be opened
 *					fHandle		- Location of file
 *					type		- WRITE - Create a new file or replace an existing file
 *								  READ - Read data from an existing file
 *								  APPEND - Append data to an existing file
 *                  
 * Output:			CE_GOOD				- FILEopen successful
 *					CE_NOT_INIT			- Card is not yet initialized because of some error
 *					CE_FILE_NOT_FOUND	- Could not find the file on the card
 *					CE_BAD_SECTOR_READ	- A bad read occured of a sector  
 *					
 *
 * Side Effects:    None
 *
 * Overview:        Opens a file to perform operations on
 *
 * Note:            None
 *****************************************************************************/

CETYPE FILEopen (FILEOBJ fo, WORD *fHandle, char type)
{
	DISK   *dsk;		//Disk structure
	BYTE 	r;					//Result of search for file
	DWORD 	l;					//lba of first sector of first cluster
	CETYPE 	error = CE_GOOD;

	dsk = (DISK *)(fo->dsk);
	if (dsk->mount == FALSE)
		error = CE_NOT_INIT;
	else
	{
        // load the sector
    	Cache_File_Entry(fo, fHandle, TRUE);
       
        // Fill up the File Object with the information pointed to by fHandle
        r = Fill_File_Object(fo, fHandle);
		if (r != FOUND)
			error = CE_FILE_NOT_FOUND;
		else
		{
			fo->seek = 0;               // first byte in file
			fo->ccls = fo->cluster;     // first cluster
			fo->sec = 0;                // first sector in the cluster
			fo->pos = 0;                // first byte in sector/cluster
			
    	    if  ( r == NOT_FOUND)
    	    {
    	        error = CE_FILE_NOT_FOUND;
			}
			else
			{   
			    // Determine the lba of the selected sector and load 
			    l = Cluster2Sector(dsk,fo->ccls);

				gBufferOwner = fo;
				if (lastClusterRead != l)
				{
					if ( !SECTORread( l, dsk->buffer))
						error = CE_BAD_SECTOR_READ;                         
					lastClusterRead = l;
				}
			} // -- found

			fo->Flags.FileWriteEOF = FALSE;
			// Set flag for operation type
			if (type == 'w' || type == 'a')
			{
				fo->Flags.write = 1;	//write or append
			}
			else
			{
				fo->Flags.write = 0;	//read
			} // -- flags
		} // -- r = Found
	} // -- Mounted
	return (error);
} // -- FILEopen
			
/******************************************************************************
 * Function:        CETYPE __fread ( FILE *fo, BYTE *pDest, WORD count, WORD *pBytesRead )
 *
 * PreCondition:    FILEopen in the read mode has been executed
 *
 * Input:           fo			- File to be read
 *					pDest		- Destination buffer for read bytes
 *					count		- Number of bytes to be read
 *					pBytesRead	- Number of bytes read
 *                  
 * Output:			CE_GOOD				- fread successful
 *					CE_EOF				- End of file reached
 *					
 *
 * Side Effects:    None
 *
 * Overview:        Reads data from a file into destination buffer
 *
 * Note:            None
 *****************************************************************************/

CETYPE __fread( FILE *fo, BYTE *pDest, WORD count, WORD *pBytesRead )
{
	DISK 	*dsk;					// Disk structure
	DWORD 	seek, sec_sel;
	WORD 	pos; 		//position within sector
	CETYPE	error = CE_GOOD;
	
	dsk 	= (DISK *)fo->dsk;
	pos 	= fo->pos;
	seek 	= fo->seek;
	
	*pBytesRead = 0;

	if( fo->Flags.write )
		return CE_WRITEONLY;

	// if it not my buffer, then get it from the disk.
	if( (gBufferOwner != fo) && (pos != MEDIA_SECTOR_SIZE ))
	{
		gBufferOwner = fo;
		sec_sel = Cluster2Sector(dsk,fo->ccls);
		sec_sel += (WORD)fo->sec;		// add the sector number to it

		if( !SECTORread( sec_sel, dsk->buffer) )
			return CE_BAD_SECTOR_READ;
		lastClusterRead = sec_sel;
	}

	//loop reading (count) bytes
	while( count )
	{
		if( seek == fo->size )
		{	error = CE_EOF;
			break; 	}

		// In fopen, pos is init to 0 and the sect is loaded
        if( pos == MEDIA_SECTOR_SIZE )
        {
            // reset position
            pos = 0;

            // point to the next sector
            fo->sec++;

            // get a new cluster if necessary
            if( fo->sec == dsk->SecPerClus )
            {
                fo->sec = 0;
                if( (error = FILEget_next_cluster( fo, 1)) != CE_GOOD )
					break;
            }

            sec_sel = Cluster2Sector(dsk,fo->ccls);
			sec_sel += (WORD)fo->sec;		// add the sector number to it

	        gBufferOwner = fo;
            if( !SECTORread( sec_sel, dsk->buffer) )
			{	error = CE_BAD_SECTOR_READ;
				break; }
			lastClusterRead = sec_sel;
        }

        // copy one byte at a time
        *pDest = RAMread( dsk->buffer, pos++ );
		pDest++;
        seek++;
		(*pBytesRead)++;
        count--;
	}

	// save off the positon
	fo->pos = pos;

	// save off the seek
	fo->seek = seek;

    return(error);
} // fread

/******************************************************************************
 * Function:        BYTE FILEget_next_cluster(FILEOBJ fo, WORD n)
 *
 * PreCondition:    Disk mounted, fo contains valid file data
 *
 * Input:           fo		- File structure
 *					n		- number of links in the FAT cluster chain to jump
 *							  through; n == 1 - next cluster in the chain
 *                  
 * Output:			CE_GOOD				- Operation successful
 *					CE_BAD_SECTOR_READ	- A bad read occured of a sector 
 *					CE_INVALID_CLUSTER	- Invalid cluster value > maxcls 
 *					CE_FAT_EOF			- Fat attempt to read beyond EOF
 *					
 *
 * Side Effects:    None
 *
 * Overview:        Steps through a chain of clusters
 *
 * Note:            None
 *****************************************************************************/

BYTE FILEget_next_cluster(FILEOBJ fo, WORD n)
{
    WORD         c, c2;
    BYTE          error = CE_GOOD;
    DISK *      disk;
   
    disk = fo->dsk;     
        
    // loop n times
    do {    
        // get the next cluster link from FAT
        c2 = fo->ccls;
        
        if ( (c = FATread( disk, c2)) == FAIL)
            error = CE_BAD_SECTOR_READ;
        else
        {    
            // check if cluster value is valid
            if ( c >= disk->maxcls)
            {
                error = CE_INVALID_CLUSTER;
            }
            
            // compare against max value of a cluster in FAT
            c2 = LAST_CLUSTER;
            
            // return if eof
            if ( c >= c2)    // check against eof
            {
                error = CE_FAT_EOF;
            }
        }
        
        // update the FILE structure
        fo->ccls = c;

    } while (--n > 0 && error == CE_GOOD);// loop end
        

    return(error);
} // get next cluster


/******************************************************************************
 * Function:        BYTE Write_File_Entry( FILEOBJ fo, WORD * curEntry)
 *
 * PreCondition:    Disk mounted, fo contains valid file data
 *
 * Input:           fo			- File structure
 *					curEntry	- Write destination
 *                  
 * Output:			TRUE	- Operation successful
 *					FALSE	- Operation failed 
 *
 * Side Effects:    None
 *
 * Overview:        Write the buffer into the current entry
 *
 * Note:            None
 *****************************************************************************/

BYTE Write_File_Entry( FILEOBJ fo, WORD * curEntry)
{
    DISK	*dsk;
    BYTE	status;
	BYTE	offset2;
    DWORD	sector;
	WORD	ccls;
    
    dsk = fo->dsk;
    
    // get the cluster of this entry
    ccls = fo->dirccls;
      
    offset2  = (*curEntry >> 4);
      
    // if its not the root, it works cluster based  
    if(ccls != 0)
        offset2 = offset2 % (dsk->SecPerClus);
        
	sector = Cluster2Sector(dsk,ccls);
              
	// Now write it    	
	if ( !SECTORwrite( sector + offset2, dsk->buffer)) 
	    status = FALSE;
	else
	    status = TRUE;                      
    
    return(status);
} // Write_File_Entry 



/******************************************************************************
 * Function:        BYTE FAT_erase_cluster_chain (WORD cluster, DISK * dsk)
 *
 * PreCondition:    Disk mounted, should be called from FILEerase
 *
 * Input:           cluster	- The cluster number
 *					dsk		- The disk structure
 *                  
 * Output:			TRUE		- Operation successful
 *					FALSE		- Operation failed 
 *					
 *
 * Side Effects:    None
 *
 * Overview:        Erase the cluster chain
 *
 * Note:            None
 *****************************************************************************/

BYTE FAT_erase_cluster_chain (WORD cluster, DISK * dsk)
{ 
    WORD     c,c2;
    enum    _status {Good, Fail, Exit}status;
    
    status = Good;
    
    // Make sure there is actually a cluster assigned
    if(cluster == 0 || cluster == 1)
    {
        status = Exit;
    }
    else
    {
        while(status == Good)
        {
            // Get the FAT entry
            if((c = FATread( dsk, cluster)) == FAIL)
                status = Fail;
            else
            {    
                if(c == 0 || c == 1)
				{
					 status = Exit;			
				}
				else
				{
					// compare against max value of a cluster in FATxx
	                c2 = LAST_CLUSTER;
	                
	                // look for the last cluster in the chain
	                if ( c >= c2)    
	                    status = Exit;
	                
	                // Now erase this FAT entry
	                if(FATwrite(dsk, cluster, CLUSTER_EMPTY, FALSE) == FAIL)
	                    status = Fail;
	                
	                // now update what the current cluster is 
	                cluster = c;
				}
            }
        }// while status
    }// cluster == 0
    
	FATwrite (dsk, 0, 0, TRUE);

    if(status == Exit)
        return(TRUE);
    else    
        return(FALSE);
} // Erase cluster


/******************************************************************************
 * Function:        DIRENTRY Cache_File_Entry( FILEOBJ fo, WORD * curEntry, BYTE ForceRead)
 *
 * PreCondition:    Disk mounted
 *
 * Input:           fo			- File information
 *					curEntry	- Location of the sector
 *					ForceRead	- Forces loading of a new sector of the root
 *                  
 * Output:			DIRENTRY	- Sector loaded
 *
 * Side Effects:    None
 *
 * Overview:        Get the sector loaded as pointed to by curEntry
 *
 * Note:            Should not be called by user
 *****************************************************************************/

DIRENTRY Cache_File_Entry( FILEOBJ fo, WORD * curEntry, BYTE ForceRead)
{
    DIRENTRY dir;
    DISK *dsk;
    DWORD sector;
    WORD cluster;
    WORD ccls;
    BYTE offset2;
    BYTE numofclus;

    dsk = fo->dsk;
    
    // get the base sector of this directory
    cluster = fo->dirclus;
    ccls = fo->dirccls;

	// figure out the offset from the base sector
	offset2  = (*curEntry >> 4);
	offset2 = offset2; // emulator issue
          
	// if its the root its not cluster based
	if(cluster != 0)  
	    offset2  = offset2 % (dsk->SecPerClus);   // figure out the offset
             
    // check if a new sector of the root must be loaded
    if (ForceRead || (*curEntry & 0xf) == 0)    // only 16 entries per sector
    {
        // see if we have to load a new cluster
        if((offset2 == 0 && (*curEntry) > DIRENTRIES_PER_SECTOR) || ForceRead)
        {
			if(cluster == 0)
			{
			    ccls = 0;
			}
			else
			{
    			if(ForceRead)
    				numofclus = ((WORD)(*curEntry) / (WORD)(((WORD)DIRENTRIES_PER_SECTOR) * (WORD)dsk->SecPerClus));
    			else
    				numofclus = 1;
    
    			// move to the correct cluster
    			while(numofclus)
    			{
    				ccls = FATread(dsk, ccls);  
    				
    				if(ccls >= LAST_CLUSTER)
    				    break;
    				else          
    				    numofclus--;
    			}
    		}	
        }       
        
        // see if that we have a valid cluster number
        if(ccls < LAST_CLUSTER)
        {           
			fo->dirccls = ccls; // write it back
            
            sector = Cluster2Sector(dsk,ccls);

            // see if we are root and about to go pass our boundaries
            if(ccls == 0 && (sector + offset2) >= dsk->data)
            {
                dir = ((DIRENTRY)NULL);   // reached the end of the root!
            }
            else
            {
                gBufferOwner = NULL;

                if ( SECTORread( sector + offset2, dsk->buffer) != TRUE) 
				{
					dir = ((DIRENTRY)NULL);
				}
				else
				{
	       		    if(ForceRead)
	       		        dir = (DIRENTRY)((DIRENTRY)dsk->buffer) + ((*curEntry)%DIRENTRIES_PER_SECTOR);
	                   else
	                       dir = (DIRENTRY)dsk->buffer;                        
				}						
				lastClusterRead = 0xFFFFFFFF;
            }
        }
        else
             dir = ((DIRENTRY)NULL);   
    }
    else
        dir = (DIRENTRY)((DIRENTRY)dsk->buffer) + ((*curEntry)%DIRENTRIES_PER_SECTOR);

    return(dir);
} // Cache_File_Entry 


/******************************************************************************
 * Function:        CETYPE __fwrite( FILE* fo, void * src, WORD count, WORD* pBytesWritten )
 *
 * PreCondition:    File opened in WRITE mode
 *
 * Input:           fo			- Pointer to file structure
 *					src			- Pointer to source buffer
 *					count		- Number of bytes to transfer
 *					pBytesWritten	- The number of bytes that have been written
 *                  
 * Output:			CE_GOOD			- Write successful
 *					CE_WRITE_ERROR	- Write failed
 *					Default			- Number of bytes that could not be written
 *
 * Side Effects:    None
 *
 * Overview:        Write file
 *
 * Note:            None
 *****************************************************************************/

CETYPE __fwrite( FILE* fo, void * src, WORD count, WORD* pBytesWritten )
{
    DISK *      dsk;     				// pointer to disk structure
	CETYPE      error = CE_GOOD;
	BYTE        sectorloaded = FALSE;
    WORD        pos;
    DWORD       l;               		// absolute lba of sector to load
    DWORD       seek,size;
//	WORD		tempo;
//	WORD 		test;
   
    // see if the file was openned in a write mode 
    if(!(fo->Flags.write))
		return CE_WRITE_ERROR;
	if (WriteProtectState())
		return CE_WRITE_PROTECTED;

  	dsk = fo->dsk;   
	        
	// get the stated position
	pos = fo->pos;
	seek = fo->seek;

	l = Cluster2Sector(dsk,fo->ccls);     
	l += (WORD)fo->sec;		// add the sector number to it
		
	gBufferOwner = fo;
	if (lastClusterRead != l)
	{
	   	if(!SECTORread( l, dsk->buffer) )
			error = CE_BAD_SECTOR_READ;
		lastClusterRead = l;
	}	

	sectorloaded = TRUE;
	     
		    // exit loop if EOF reached 
	size = fo->size;
	        
	        // 2. loop reading count bytes 
	while (error == CE_GOOD && count > 0)
	{          
		if( seek == size )
			fo->Flags.FileWriteEOF = TRUE;

		// load a new sector if necessary, multiples of sector
		if (pos == MEDIA_SECTOR_SIZE) 
		{    
			// close the currently loaded sector if we need to	
			if(sectorloaded)                                
			{
	    		if(SECTORwrite( l, dsk->buffer) != TRUE)
					error = CE_WRITE_ERROR; 
			}    
		            
		    // reset position
		    pos = 0;
		            
	        // point to the next sector
		    fo->sec++;
		            
		    // get a new cluster if necessary            
		    if (fo->sec == dsk->SecPerClus)
		    {
		        fo->sec = 0;

				if(fo->Flags.FileWriteEOF)
					error = FILEallocate_new_cluster(fo);    // add new cluster to the file
				else
					error = FILEget_next_cluster( fo, 1);			        
		    }
		    		
			if (error == CE_DISK_FULL)
				return error;		           
				
			if(error == CE_GOOD)
		    {
				l = Cluster2Sector(dsk,fo->ccls);
				l += (WORD)fo->sec;		// add the sector number to it
				gBufferOwner = fo;
				if( !SECTORread( l, dsk->buffer) )
				{
					error = CE_BAD_SECTOR_READ;
					lastClusterRead = 0xFFFFFFFF;
				}
				else
				{
					lastClusterRead = l;
					sectorloaded = TRUE;
				}
			}
		} //  load new sector
	
		if(error == CE_GOOD)
		{
			// Write one byte at a time
			RAMwrite(dsk->buffer, pos++, *(char *)src);
			src = src + 1; // compiler bug
			seek++;
			count--;
			(*pBytesWritten)++;
			// now increment the size of the part
			if(fo->Flags.FileWriteEOF)
	 	   	size++;             
		}
	         	
	} // while count
	        
	if(error == CE_GOOD)
	{
		// figure out the lba
	 	l = Cluster2Sector(dsk,fo->ccls);
	 			       
	    l += (WORD)fo->sec;		// add the sector number to it
	
	    if(!SECTORwrite( l, dsk->buffer))
	    	error = CE_WRITE_ERROR; 
	}
	        
	// save off the positon
	fo->pos = pos;
	    
	// save off the seek
	fo->seek = seek;
	    	
	// now the new size
	fo->size = size;

    return(error);
} // __fwrite


/******************************************************************************
 * Function:        CETYPE CreateFileEntry(FILEOBJ fo, WORD *fHandle)
 *
 * PreCondition:    File opened in WRITE mode
 *
 * Input:           fo			- Pointer to file structure
 *					fHandle		- Location to create file
 *                  
 * Output:			CE_GOOD			- File creation successful
 *					CE_DIR_FULL		- All root dir entry are taken
 *
 * Side Effects:    None
 *
 * Overview:        With the data passed within fo, create a new file entry in the current directory
 *
 * Note:            None
 *****************************************************************************/

CETYPE CreateFileEntry(FILEOBJ fo, WORD *fHandle)
{
	BYTE    index;
//	BYTE 	size;
	CETYPE  error = CE_GOOD;
//	BYTE    numdirentries;
   	char    name[11];
 
	for (index = 0; index < FILE_NAME_SIZE; index ++)
	{
		name[index] = fo->name[index];
	}

	// make sure we are good
	if(error == CE_GOOD)
	{ 
	    *fHandle = 0;
	    
	    // figure out where to put this file in the directory stucture 
	    if(FindEmptyEntries(fo, fHandle))
	    {	    
    		// found the entry now populate it 
            if((error = PopulateEntries(fo, name ,fHandle)) == CE_GOOD)
            {
                // if everything is a go, create a first cluster
				error = CreateFirstCluster(fo);                
            }
        }
        else
        {
            error = CE_DIR_FULL;
        }		
	}
	return(error);
}

/******************************************************************************
 * Function:        CETYPE CreateFirstCluster(FILEOBJ fo)
 *
 * PreCondition:    Called from CreateFileEntry
 *
 * Input:           fo	- The file that contains the first cluster
 *                  
 * Output:			CE_GOOD			- First cluster created
 *					CE_WRITE_ERROR	- Cluster creation failed
 *
 * Side Effects:    None
 *
 * Overview:        Finds an open cluster and links it in
 *
 * Note:            Should not be called by user
 *****************************************************************************/

CETYPE CreateFirstCluster(FILEOBJ fo)
{
    CETYPE    error;
    WORD      cluster;
    WORD      fHandle;
    DIRENTRY  dir;
    
    fHandle = fo->entry;
    
	// Now create the first cluster (head cluster)
	if((error = FILECreateHeadCluster(fo,&cluster)) == CE_GOOD)
    {        
        // load it so I can link in the new cluster
        dir = LoadDirAttrib(fo, &fHandle);

        // Now update the new cluster
        dir->DIR_FstClusLO = cluster;

        // now write it
        if(Write_File_Entry(fo, &fHandle) != TRUE)
            error = CE_WRITE_ERROR;
    } // Create Cluster   
    
    return(error);
}// End of CreateFirstCluster

/******************************************************************************
 * Function:        BYTE FindEmptyEntries(FILEOBJ fo, WORD *fHandle)
 *
 * PreCondition:    Disk mounted, CreateFileEntry called
 *
 * Input:           fo			- Pointer to file structure
 *					fHandle		- Start of entries
 *                  
 * Output:			TRUE	- One found
 *					FALSE	- None found
 *
 * Side Effects:    None
 *
 * Overview:        Find the passed number of contingiant empty entries
 *
 * Note:            Should not be called by user
 *****************************************************************************/

BYTE FindEmptyEntries(FILEOBJ fo, WORD *fHandle)
{
    BYTE	status = NOT_FOUND;
    BYTE	amountfound;
    BYTE	a;
    WORD	bHandle;
    DIRENTRY    dir;

    if((dir = Cache_File_Entry( fo, fHandle, TRUE)) == NULL)
    {
        status = CE_BADCACHEREAD;
    }
    else
    {
        // while its still not found
        while(status == NOT_FOUND)
        {
            amountfound = 0;
            bHandle = *fHandle;
            
            // find (number) continuous entries
            do
            {
                // Get the entry
                dir = Cache_File_Entry( fo, fHandle, FALSE);
                
                // Read the first char of the file name 
                a = dir->DIR_Name[0];     
                
                // increase number
                (*fHandle)++;
            }while((a == DIR_DEL || a == DIR_EMPTY) && (dir != (DIRENTRY)NULL) &&  (++amountfound < 1));  
            
            // --- now why did we exit?
            if(dir == NULL) // Last entry of the cluster
            {
                //setup the current cluster
                a = fo->dirccls; // write it back
                
                // make sure we are not the root directory
                if(a == 0)
                    status = NO_MORE;
                else    
                {
                    fo->ccls = a;
                    
                    if(FILEallocate_new_cluster(fo) == CE_DISK_FULL)
                        status = NO_MORE;    
                    else
                    {
                        *fHandle = bHandle;
                        status = FOUND;     // a new cluster will surely hold a new file name     
                    }
                }            
            }
            else
            {            
                if(amountfound == 1)
                {
                    status = FOUND;
                    *fHandle = bHandle;
                }
            }
                
            
    	}// while    
        
    	// copy the base handle over
    	*fHandle = bHandle;
    }

    if(status == FOUND)
        return(TRUE);
    else
        return(FALSE);        
}

/******************************************************************************
 * Function:        BYTE PopulateEntries(FILEOBJ fo, char *name , WORD *fHandle)
 *
 * PreCondition:    Disk mounted, CreateFileEntry called
 *
 * Input:           fo			- Pointer to file structure
 *					name		- Name of the file
 *					fHandle		- Location of the file
 *                  
 * Output:			CE_GOOD		- Population successful
 *
 * Side Effects:    None
 *
 * Overview:        Actually fill up the directory entry with the data at hand and write it  
 *
 * Note:            Should not be called by user
 *****************************************************************************/

BYTE PopulateEntries(FILEOBJ fo, char *name , WORD *fHandle)
{
    BYTE error = CE_GOOD;
//    BYTE nameptr, temp, index, csum;
    DIRENTRY    dir;

    dir = Cache_File_Entry( fo, fHandle, TRUE);
        
    // copy the contents over
    strncpy(dir->DIR_Name,name,DIR_NAMECOMP);
    	
    // setup no attributes
    dir->DIR_Attr   = ATTR_ARCHIVE;
    dir->DIR_NTRes  = 0x00; // nt reserved
    
    // we don't have a real time clock
    dir->DIR_CrtTimeTenth = 0xB2; // millisecond stamp
    dir->DIR_CrtTime =      0x7278;      // time created //
    dir->DIR_CrtDate =      0x32B0;      // date created (1/1/2004)
    dir->DIR_LstAccDate =   0x32B0;   // Last Access date
    dir->DIR_FstClusHI =    0x00;    // high word of this enty's first cluster number
    dir->DIR_WrtTime =      0x7279;      // last update time
    dir->DIR_WrtDate =      0x32B0;      // last update date
    dir->DIR_FstClusLO =    0x0000;    // low word of this entry's first cluster number
    dir->DIR_FileSize =     0x0;     // file size in DWORD    
    
   	fo->size        = dir->DIR_FileSize;
	fo->time        = dir->DIR_CrtTime;
	fo->date        = dir->DIR_CrtDate;
	fo->attributes  = dir->DIR_Attr;
    fo->entry       = *fHandle;
    
    // just write the last entry in
    Write_File_Entry(fo,fHandle);
    
    return(error);
}

/******************************************************************************
 * Function:        BYTE FILEallocate_new_cluster( FILEOBJ fo)
 *
 * PreCondition:    Disk mounted
 *
 * Input:           fo		- Pointer to file structure
 *                  
 * Output:			OK		- Cluster allocated
 *
 * Side Effects:    None
 *
 * Overview:        Allocate new cluster
 *
 * Note:            Should not be called by user
 *****************************************************************************/

BYTE FILEallocate_new_cluster( FILEOBJ fo)
// fo   file structure
{
    DISK *      dsk;
    WORD c,curcls;
    
    dsk = fo->dsk;   
    c = fo->ccls;
        
    // find the next empty cluster
    c = FATfindEmptyCluster(fo);
    if (c == 0)
		return CE_DISK_FULL;
    // mark the cluster as taken, and last in chain
    if(dsk->type == FAT12) 
        FATwrite( dsk, c, LAST_CLUSTER_FAT12, FALSE);
    else
        FATwrite( dsk, c, LAST_CLUSTER_FAT16, FALSE);
    
    // link current cluster to the new one
    curcls = fo->ccls;
    
    FATwrite( dsk, curcls, c, FALSE);
    
    // update the FILE structure
    fo->ccls = c;
        
    return CE_GOOD;
} // allocate new cluster


/******************************************************************************
 * Function:        WORD FATfindEmptyCluster(FILEOBJ fo)
 *
 * PreCondition:    Disk mounted
 *
 * Input:           fo		- Pointer to file structure
 *                  
 * Output:			c		- Cluster; 0 if failed
 *
 * Side Effects:    None
 *
 * Overview:        Find the next available cluster
 *
 * Note:            Should not be called by user
 *****************************************************************************/

WORD FATfindEmptyCluster(FILEOBJ fo)
{    
    DISK *      disk;
    WORD         value=0x0,c,curcls;
    
    disk = fo->dsk;   
    c = fo->ccls;
    
    // just incase
    if(c < 2)
        c = 2;

    curcls = c;
        
    FATread(disk, c);
        
    // sequentially scan through the FAT looking for an empty cluster
    while(c)
    {
        c++;    // check next cluster in FAT
        // check if reached last cluster in FAT, re-start from top
        if (value == END_CLUSTER || c >= disk->maxcls)
            c = 2;  
    
        // check if full circle done, disk full        
        if ( c == curcls)
        {
            c = 0;
            break;
        }
            
        // look at its value 

        if ( (value = FATread(disk, c)) == CLUSTER_FAIL)
        {
            c = 0;
            break;
        }

        // check if empty cluster found
        if (value == CLUSTER_EMPTY)
            break;
            
    }  // scanning for an empty cluster
    
    return(c);
}






/******************************************************************************
 * Function:        CETYPE __fclose(FILEOBJ fo)
 *
 * PreCondition:    File opened
 *
 * Input:           fo		- Pointer to file structure
 *                  
 * Output:			CE_GOOD				- File closed successfully
 *					CE_WRITE_ERROR		- Could not write to the sector
 *					CE_FILENOTOPENED	- File not opened for the write 
 *
 * Side Effects:    None
 *
 * Overview:        Close a file
 *
 * Note:
 *****************************************************************************/

CETYPE __fclose(FILEOBJ fo)
{
    WORD        fHandle, fIndex;
    CETYPE        error = 72;
    DIRENTRY    dir;

    fHandle = fo->entry;

    if(fo->Flags.write)
    {
		// Write the current FAT sector to the disk
		FATwrite (fo->dsk, 0, 0, TRUE);

        // Get the file entry
        dir = LoadDirAttrib(fo, &fHandle);

		if (dir == NULL)
		{
			error = CE_DIR_NOT_FOUND;
			return error;
		}

        // update the time
        IncrementTimeStamp(dir);

		Nop();

        dir->DIR_FileSize = fo->size;

        // just write the last entry in
        if(Write_File_Entry(fo,&fHandle))
            error = CE_GOOD;
        else
            error = CE_WRITE_ERROR;

        // its now closed!
        fo->Flags.write = FALSE;
    }

#ifdef FAT16_DYNAMIC_MEM
	FAT16_free((unsigned char *)fo);
#else

	for( fIndex = 0; fIndex < FAT16_MAX_FILES_OPEN; fIndex++ )
	{
		if( fo == &gFileArray[fIndex] )
		{
			gFileSlotOpen[fIndex] = TRUE;
			break;
		}
	}


#endif

    if (error == 72)
		error = CE_GOOD;

    return(error);
} // fclose



/******************************************************************************
 * Function:        void IncrementTimeStamp(DIRENTRY dir)
 *
 * PreCondition:    File opened
 *
 * Input:           dir		- Pointer to directory structure
 *                  
 * Output:			void
 *
 * Side Effects:    None
 *
 * Overview:        Touch the time so we can keep track of the files
 *
 * Note:            None
 *****************************************************************************/

void IncrementTimeStamp(DIRENTRY dir)
{
    BYTE          seconds;
    BYTE          minutes;
    BYTE          hours;
    
    BYTE          day;
    BYTE          month;
    BYTE          year;
    
    seconds = (dir->DIR_WrtTime & 0x1f);
    minutes = ((dir->DIR_WrtTime & 0x07E0) >> 5);
    hours   = ((dir->DIR_WrtTime & 0xF800) >> 11);

    day     = (dir->DIR_WrtDate & 0x1f);
    month   = ((dir->DIR_WrtDate & 0x01E0) >> 5);
    year    = ((dir->DIR_WrtDate & 0xFE00) >> 9);
        
    if(seconds < 29)
        seconds++;
    else
    {
        seconds = 0x00;
        
        if(minutes < 59)
        {
           minutes++; 
        }
        else
        {
            minutes = 0;
            
            if(hours < 23)
            {
                hours++;
            }
            else
            {
                hours = 0;
                // new day!... only 28 in a month
                if(day < 28)
                {
                    day++;
                }
                else
                {
                    day = 1;
                    
                    if(month < 12)
                    {
                        month++;
                    }
                    else
                    {
                        month = 1;
                        // new year
                        year++;
                        // HOPEFULLY we never go past 2107
                    }
                }
            }
        }
    }
    
    dir->DIR_WrtTime = (WORD)(seconds);
    dir->DIR_WrtTime |= ((WORD)(minutes) << 5);
    dir->DIR_WrtTime |= ((WORD)(hours) << 11);  
    
    dir->DIR_WrtDate = (WORD)(day);
    dir->DIR_WrtDate |= ((WORD)(month) << 5);
    dir->DIR_WrtDate |= ((WORD)(year) << 9);
}


/******************************************************************************
 * Function:        BYTE Fill_File_Object(FILEOBJ fo, WORD *fHandle)
 *
 * PreCondition:    Disk mounted
 *
 * Input:           fo		- Pointer to file structure
 *					fHandle - Passed member's location
 *                  
 * Output:			FOUND		- Operation successful
 *					NOT_FOUND	- Operation failed
 *
 * Side Effects:    None
 *
 * Overview:        Fill the passed file object with the passed member's (fHandle) information
 *
 * Note:            Should not be called by user
 *****************************************************************************/

BYTE Fill_File_Object(FILEOBJ fo, WORD *fHandle)
{
    DIRENTRY    dir;
    BYTE 			index, a;         
    BYTE     character;
    BYTE          status;   
    DWORD         temp;
	BYTE	test = 0;
    
    // Get the entry
    dir = Cache_File_Entry( fo, fHandle, FALSE);
    
    // Read the first char of the file name 
    a = dir->DIR_Name[0]; 
	    
    // Make sure there is a directory left
    if(dir == (DIRENTRY)NULL || a == DIR_EMPTY)
    {                  
        status = NO_MORE;
    }
    else
    {
        // Check for empty or deleted directory
        if ( a == DIR_DEL)
            status = NOT_FOUND;
        else
        {
            // Get the attributes
            a = dir->DIR_Attr; 
            

			/* Short File Name */
        
			// 2.1.1  print the file name and extension         					
			for (index=0; index < DIR_NAMESIZE; index++)
			{
				character = dir->DIR_Name[index];
				character = (BYTE)toupper(character); 

				fo->name[test++] = character;
			}
			
			// Get the attributes
			a = dir->DIR_Attr; 

			// its possible to have an extension in a directory
			character = dir->DIR_Extension[0];


			// Get the file extension if its there
			for (index=0; index < DIR_EXTENSION; index++)
			{
				character = dir->DIR_Extension[index];
				character = (BYTE)toupper(character);


				fo->name[test++] = character;
			}

			// done and done with the name
//			fo->name[++test] = (BYTE)'\0';

			// Now store the identifier
			fo->entry = *fHandle;
					

            // see if we are still a good file (long file names can erase just the main entry)
            a = dir->DIR_Name[0];
            
            if(a == DIR_DEL)
                status = NOT_FOUND;
            else
                status = FOUND;

			// Now store the size
            fo->size = (dir->DIR_FileSize);
            	
	        // Now store the cluster
	        temp = (dir->DIR_FstClusHI);
			temp = temp << 16;
	        temp |= dir->DIR_FstClusLO;
	        fo->cluster = temp;
	                  
	        // get the date and time      
	        fo->time = (dir->DIR_WrtTime);
            fo->date = (dir->DIR_WrtDate);
	
	        /// -Get and store the attributes
            a = dir->DIR_Attr;           
            fo->attributes = a;
 
        }// deleted directory  
    }// Ensure we are still good
        
    return(status);
} // Fill_File_Object


/******************************************************************************
 * Function:        DIRENTRY LoadDirAttrib(FILEOBJ fo, WORD *fHandle)
 *
 * PreCondition:    Disk mounted
 *
 * Input:           fo		- Pointer to file structure
 *					fHandle	- Information location
 *                  
 * Output:			DIRENTRY	- The directory entry
 *
 * Side Effects:    None
 *
 * Overview:        Load the current sector for the filename base and return the direntry
 *
 * Note:            Should not be called by user
 *****************************************************************************/

DIRENTRY LoadDirAttrib(FILEOBJ fo, WORD *fHandle)
{
    DIRENTRY    dir;
    BYTE		a;         
    
    // Get the entry
    dir = Cache_File_Entry( fo, fHandle, TRUE);
    
    // Read the first char of the file name 
    a = dir->DIR_Name[0]; 
	    
    // Make sure there is a directory left
    if(a == DIR_EMPTY)
        dir = (DIRENTRY)NULL;
    
    if(dir != (DIRENTRY)NULL)
    {
        // Check for empty or deleted directory
        if ( a == DIR_DEL)
            dir = (DIRENTRY)NULL;
        else
        {
            // Get the attributes
            a = dir->DIR_Attr; 
            
             // scan through all the long dir entries
            while(a == ATTR_LONG_NAME)
            {   
                (*fHandle)++;
                dir = Cache_File_Entry( fo, fHandle, FALSE);
                a = dir->DIR_Attr;      
            } // long file name while loop
            
        } // deleted dir
        
    }// Ensure we are still good
        
    return(dir);
} // LoadDirAttrib


/******************************************************************************
 * Function:        CETYPE FILEerase( FILEOBJ fo, WORD *fHandle, BYTE EraseClusters)
 *
 * PreCondition:    File not opened, File exists
 *
 * Input:           fo				- Pointer to file structure
 *					fHandle 		- Location of file information
 *					EraseClusters 	- Remove cluster allocation from FAT?
 *                  
 * Output:			CE_GOOD				- File erased successfully
 *					CE_FILE_NOT_FOUND	- Could not find the file on the card
 *					CE_ERASE_FAIL		- Internal Card erase failed
 *
 * Side Effects:    None
 *
 * Overview:        Erase file
 *
 * Note:            Should not be called by user
 *****************************************************************************/

CETYPE FILEerase( FILEOBJ fo, WORD *fHandle, BYTE EraseClusters)
{
    DIRENTRY    dir;
    BYTE		a;           
    CETYPE      status = CE_GOOD;   
    WORD        clus;  
    DISK *      disk;
    
    disk = fo->dsk;    

    // reset the cluster
    clus = fo->dirclus;
    fo->dirccls = clus;
    
    // load the sector
    dir = Cache_File_Entry(fo, fHandle, TRUE);
    
    // Fill up the File Object with the information pointed to by fHandle
    a = dir->DIR_Name[0]; 
	    
    // see if there is something in the dir
    if(dir == (DIRENTRY)NULL || a == DIR_EMPTY)
    {                  
        status = CE_FILE_NOT_FOUND;
    }
    else
    {
        // Check for empty or deleted directory
        if ( a == DIR_DEL)
            status = CE_FILE_NOT_FOUND;
        else
        {        
            // Get the attributes
            a = dir->DIR_Attr; 
              
		    /* Short File Name - entry*/
		    dir->DIR_Name[0] = DIR_DEL; // mark as deleted
	
	        // Get the starting cluster
	        clus = dir->DIR_FstClusLO;
	        
            // Now write it 
            if(status != CE_GOOD || !(Write_File_Entry( fo, fHandle)))
                status = CE_ERASE_FAIL;
            else
            {
                if(EraseClusters)
                {
                    /* Now remove the cluster allocation from the FAT */
                    status = ((FAT_erase_cluster_chain(clus, disk)) ? CE_GOOD : CE_ERASE_FAIL);    
                }
            }
        
        } // Not already deleted
        
    }// Not existant
            
    return (status);
}

/******************************************************************************
 * Function:        FILE * __fopen (const char * fileName, const char *mode, FILE* filePtr )
 *
 * PreCondition:    For read or append mode, file exists
 *
 * Input:           fileName	- The name of the file to open
*					mode	- WRITE - Create a new file or replace an existing file
 *							  READ - Read data from an existing file
 *							  APPEND - Append data to an existing file
 *					filePte	- The pointer to the file that will be returned
 *								in the parent function of __fopen
 *                  
 * Output:			FILE *	- The pointer to the file object
 *
 * Side Effects:    None
 *
 * Overview:        File open function
 *
 * Note:            None
 *****************************************************************************/
FILE * __fopen (const char * fileName, const char *mode, FILE* filePtr )
{
//	BYTE	count;
	BYTE	ModeC;
	WORD 	fHandle;
	CETYPE	final;
	FILE    gblFileTemp;

	//Read the mode character
	ModeC = mode[0];

	if (WriteProtectState())
	{
		return NULL;
	}

	//Format the source string to FAT16 lib spec.
	if( !FormatFileName(fileName, filePtr->name) )
		return NULL;	//bad filename

	filePtr->dsk = &glbDiskData;
    filePtr->cluster = 0;
    filePtr->ccls    = 0;

    // start at the root directory
  	filePtr->dirclus 	= 0;
	filePtr->dirccls 	= 0;

    // copy file object over
    FileObjectCopy(&gblFileTemp, filePtr);

    // See if the file is found
	if(FILEfind (filePtr, &gblFileTemp, 1) == CE_GOOD)
	{
        // File is Found
		switch(ModeC)
		{
			case 'w':
			case 'W':
			{
			    // File exists, we want to create a new one, remove it first
				fHandle = filePtr->entry;
				final = FILEerase(filePtr, &fHandle, TRUE);

				if (final == CE_GOOD)
				{
				    // now create a new one
					final = CreateFileEntry (filePtr, &fHandle);

					if (final == CE_GOOD)
					{
						final = FILEopen (filePtr, &fHandle, 'w');

						if (final == CE_GOOD)
						{
							final = __fseek (filePtr, 0, SEEK_END);
							if (final != CE_GOOD)
								filePtr = NULL;
						}
						else
							filePtr = NULL;
					}
					else
						filePtr = NULL;
				}
				else
					filePtr = NULL;

				break;
			}
	
			case 'A':
			case 'a':
			{
				if(filePtr->size != 0)
				{
					fHandle = filePtr->entry;

					final = FILEopen (filePtr, &fHandle, 'w');

					if (final == CE_GOOD)
					{
						final = __fseek (filePtr, 0, SEEK_END);
						if (final != CE_GOOD)
							filePtr = NULL;
					}
					else
						filePtr = NULL;
				}
				else
				{
					fHandle = filePtr->entry;
					final = FILEerase(filePtr, &fHandle, TRUE);

					if (final == CE_GOOD)
					{
					    // now create a new one
						final = CreateFileEntry (filePtr, &fHandle);

						if (final == CE_GOOD)
						{
							final = FILEopen (filePtr, &fHandle, 'w');

							if (final == CE_GOOD)
							{
								final = __fseek (filePtr, 0, SEEK_END);
								if (final != CE_GOOD)
									filePtr = NULL;
							}
							else
								filePtr = NULL;
						}
						else
							filePtr = NULL;
					}
					else
						filePtr = NULL;
				}
				break;
			}

			case 'R':
			case 'r':
			{
			    fHandle = filePtr->entry;

				final = FILEopen (filePtr, &fHandle, 'r');
				if (final != CE_GOOD)
					filePtr = NULL;

				break;
			}

			default:
				filePtr = NULL;  //indicate error condition

			break;
		}
	}
	else
	{
    	// the file was not found, reset to the default asked
    	FileObjectCopy(filePtr, &gblFileTemp);

	    // File is not Found
		if(ModeC == 'w' || ModeC == 'W')
		{
    		// use the user requested name
    		fHandle = 0;
			final = CreateFileEntry (filePtr, &fHandle);

			if (final == CE_GOOD)
			{
				final = FILEopen (filePtr, &fHandle, 'w');

				if (final == CE_GOOD)
				{
					final = __fseek (filePtr, 0, SEEK_END);
					if (final != CE_GOOD)
						filePtr = NULL;
				}
				else
					filePtr = NULL;
			}
			else
				filePtr = NULL;
		}
		else
			filePtr = NULL;
	}

   	return(filePtr);
}


/******************************************************************************
 * Function:        CETYPE __fseek (FILEOBJ fo, DWORD offset, BYTE origin)
 *
 * PreCondition:    File exists
 *
 * Input:           fo		- Pointer to file structure
 *					offset	- Offset from origin location
 *					origin	- SEEK_SET - Seek from start of file
 *							  SEEK_CUR - Seek from current location
 *							  SEEK_END - Seek from end of file
 *                  
 * Output:			CE_GOOD						- Operation successful
 *					CE_INVALID_CLUSTER			- Invalid cluster value > maxcls
 *					CE_BAD_SECTOR_READ			- A bad read occured of a sector 
 *
 * Side Effects:    None
 *
 * Overview:        User called function to set the starting position of the stream to the offset 
 *
 * Note:            None
 *****************************************************************************/

CETYPE __fseek (FILEOBJ fo, DWORD offset, BYTE origin)
{
	DWORD 	numsector, temp;	// lba of first sector of first cluster
	DISK*	dsk;				// pointer to disk structure

	dsk = fo->dsk;

	switch(origin)
	{
		case SEEK_CUR:
		// Apply the offset to the current position
			offset += fo->seek;
		break;
		case SEEK_END:
		// Apply the offset to the end of the file
			offset = fo->size - offset;
		break;
		case SEEK_SET:
		// automatically there
		default:
		break;
	}

    // start from the beginning
	temp = fo->cluster;
	fo->ccls = temp;

	temp = fo->size;

	if (offset > temp)
		return (-1);		// past the limits
	else
	{
        // if we are writing we are no longer at the end
		fo->Flags.FileWriteEOF = FALSE;

        // set the new postion
		fo->seek = offset;

        // figure out how many sectors
		numsector = offset / MEDIA_SECTOR_SIZE;

        // figure out how many bytes off of the offset
		offset = offset - (numsector * MEDIA_SECTOR_SIZE);
		fo->pos = offset;

        // figure out how many clusters
		temp = numsector / dsk->SecPerClus;

        // figure out the stranded sectors
        numsector = numsector - (dsk->SecPerClus * temp);
		fo->sec = numsector;

        // if we are in the current cluster stay there
		if (temp > 0)
		{
			if (FILEget_next_cluster(fo, temp) != CE_GOOD)
				return (CE_INVALID_CLUSTER);	// past the limits
		}

        // Determine the lba of the selected sector and load
		temp = Cluster2Sector(dsk,fo->ccls);

        // now the extra sectors
		numsector = fo->sec;
		temp += numsector;

		gBufferOwner = NULL;
		if( !SECTORread(temp, dsk->buffer) )
			return (CE_BAD_SECTOR_READ);	// Bad read
		lastClusterRead = temp;
	}

	return (CE_GOOD);
}

/******************************************************************************
 * Function:        long ftell (FILEOBJ fo)
 *
 * PreCondition:    File opened
 *
 * Input:           fo		- Pointer to file structure
 *                  
 * Output:			fo->seek	- Current location of the file
 *
 * Side Effects:    None
 *
 * Overview:        User called function to determine the current location of a file
 *
 * Note:            None
 *****************************************************************************/

long ftell (FILE * fo)
{
	return (fo->seek);
}

/******************************************************************************
 * Function:        int remove (const char * fileName)
 *
 * PreCondition:    File not opened, file exists
 *
 * Input:           compare	- Pointer to file to be erased
 *                  
 * Output:			0	- File removed
 *					-1	- File was not removed
 *
 * Side Effects:    None
 *
 * Overview:        User called function to erase a file
 *
 * Note:            None
 *****************************************************************************/

int remove (const char * fileName)
{
	FILE f;
	FILE gblFileTemp;
	FILEOBJ fo = &f;
//	BYTE count;
	CETYPE result;
	
	if (WriteProtectState())
	{
		return (-1);
	}

	//Format the source string to FAT16 lib spec.
	if( !FormatFileName(fileName, fo->name) )
		return -1;

	fo->dsk = &glbDiskData;
	fo->cluster = 0;
	fo->ccls    = 0;

    // start at the root directory
  	fo->dirclus 	= 0;
	fo->dirccls 	= 0;

    // copy file object over
    FileObjectCopy(&gblFileTemp, fo);
    
    // See if the file is found
	result = FILEfind (fo, &gblFileTemp, 1);
	
	if (result != CE_GOOD)
		return -1;
	result = FILEerase(fo, &fo->entry, TRUE);
	if( result == CE_GOOD )
		return 0;
	else
		return -1;
}

/******************************************************************************
 * Function:        int fseek(FILE *stream, long offset, int whence)
 *
 * PreCondition:    File exists
 *
 * Input:           stream	- Pointer to file structure
 *					offset	- Offset from origin location
 *					whence	- SEEK_SET - Seek from start of file
 *							  SEEK_CUR - Seek from current location
 *							  SEEK_END - Seek from end of file
 *
 * Output:			0					- Operation successful
 *					-1					- Operation unsuccesful
 *
 * Side Effects:    None
 *
 * Overview:        User called function to set the starting position of the stream to the offset
 *
 * Note:            None
 *****************************************************************************/
int fseek(FILE *stream, long offset, int whence)
{
	CETYPE err;

	err = __fseek (stream, offset, (BYTE)whence);

	if( err == CE_GOOD )
		return 0;
	else
		return -1;
}

/******************************************************************************
 * Function:        void rewind (FILEOBJ fo)
 *
 * PreCondition:    File opened, file exists
 *
 * Input:           fo		- Pointer to file structure
 *                  
 * Output:			void
 *
 * Side Effects:    None
 *
 * Overview:        User called function to reset the position of the file
 *
 * Note:            None
 *****************************************************************************/

void rewind (FILEOBJ fo)
{
	fo->seek = 0;
	fo->pos = 0;
	fo->sec = 0;
	fo->ccls = fo->cluster;
	gBufferOwner = NULL;
	return;
}

/******************************************************************************
 * Function:        void FileObjectCopy(FILEOBJ foDest,FILEOBJ foSource)
 *
 *
 * Input:			foDest		- The destination
 *					foSource	- the source
 *                  
 * Output:			None
 *
 * Side Effects:    None
 *
 * Overview:        Makes a replica of an existing file object
 *
 * Note:            None
 *****************************************************************************/

void FileObjectCopy(FILEOBJ foDest,FILEOBJ foSource)
{
    BYTE i;
    BYTE size;
    BYTE *dest;
    BYTE *source;
    
    dest = (BYTE *)foDest;
    source = (BYTE *)foSource;
    
    size = sizeof(FILE);
    
    for(i=0; i< size; i++)
    {
        dest[i] = source[i];    
    }
}

/******************************************************************************
 * Function:        CETYPE FILECreateHeadCluster( FILEOBJ fo, WORD *cluster)
 *
 * PreCondition:    Disk mounted
 *
 * Input:           fo		- Pointer to file structure
 *					cluster	- Cluster location
 *                  
 * Output:			CE_GOOD				- File closed successfully
 *					CE_WRITE_ERROR		- Could not write to the sector
 *					CE_DISK_FULL		- All clusters in partition are taken
 *
 * Side Effects:    None
 *
 * Overview:        Use to create the HEAD cluster (1st one)
 *
 * Note:            Should not be called by user
 *****************************************************************************/

CETYPE FILECreateHeadCluster( FILEOBJ fo, WORD *cluster)
{
    DISK *      disk;
//    WORD        curcls;
    CETYPE        error = CE_GOOD;
    
    disk = fo->dsk;   
    
    // find the next empty cluster
    *cluster = FATfindEmptyCluster(fo);
  
  	if(*cluster == 0)
        error = CE_DISK_FULL;
    else
    {    
        // mark the cluster as taken, and last in chain
        if(disk->type == FAT12) 
		{
            if(FATwrite( disk, *cluster, LAST_CLUSTER_FAT12, FALSE) == FAIL)
                error = CE_WRITE_ERROR;
		}
        else
		{        
		    if(FATwrite( disk, *cluster, LAST_CLUSTER_FAT16, FALSE) == FAIL)
                error = CE_WRITE_ERROR;
		}
            
        // lets erase this cluster                
        if(error == CE_GOOD)
        {
            error = EraseCluster(disk,*cluster);
        }                
    }
        
    return(error);
} // allocate head cluster

/******************************************************************************
 * Function:        BYTE EraseCluster(DISK *disk, WORD cluster)
 *
 * PreCondition:    File opened
 *
 * Input:           dsk		- Disk structure
 *					cluster	- Cluster to be erased
 *                  
 * Output:			CE_GOOD				- File closed successfully
 *					CE_WRITE_ERROR		- Could not write to the sector
 *
 * Side Effects:    None
 *
 * Overview:        Erase the passed cluster
 *
 * Note:            Should not be called by user
 *****************************************************************************/

BYTE EraseCluster(DISK *disk, WORD cluster)
{
//	BYTE NumofSectors;
	BYTE	index;
	DWORD SectorAddress;
	BYTE error = CE_GOOD;

	SectorAddress = Cluster2Sector(disk,cluster);

	gBufferOwner = NULL;

	// clear out the memory first
    memset(disk->buffer, 0x00, MEDIA_SECTOR_SIZE);
	
	// Now clear them out
	for(index = 0; index < disk->SecPerClus && error == CE_GOOD; index++)
	{
	    if (SECTORwrite( SectorAddress++, disk->buffer) != TRUE) 
	        error = CE_WRITE_ERROR;
	}
	
	return(error);
}

/******************************************************************************
 * Function:        DWORD Cluster2Sector(DISK * dsk, WORD cluster)
 *
 * PreCondition:    Disk mounted
 *
 * Input:           disk	- Disk structure
 *					cluster	- Cluster to be converted
 *                  
 * Output:			sector	- Sector that corresponds to given cluster
 *
 * Side Effects:    None
 *
 * Overview:        Given a cluster, figure out the sector number
 *
 * Note:            Should not be called by user
 *****************************************************************************/

DWORD Cluster2Sector(DISK * dsk, WORD cluster)
{
    DWORD sector;
    
	// The root dir takes up cluster 0 and 1 
	if(cluster == 0 ||cluster == 1)
		sector = dsk->root + cluster;
	else
	    sector = (((DWORD)cluster-2) * dsk->SecPerClus) + dsk->data;
    return(sector);
}

/******************************************************************************
 * Function:        size_t fwrite(const void *ptr, size_t size, size_t n, FILE *stream)
 *
 * PreCondition:    File opened in WRITE mode
 *
 * Input:           stream		- Pointer to file structure
 *					ptr			- Pointer to source buffer
 *					n			- Number of units to transfer
 *					size		- Size of units in bytes
 *
 * Output:			size_t		- number of units written
 *
 * Side Effects:    None
 *
 * Overview:        Write file
 *
 * Note:            None
 *****************************************************************************/
size_t fwrite(const void *ptr, size_t size, size_t n, FILE *stream)
{
	WORD writeCount = 0;

	if( size == 0 || n == 0 )
		return 0;

	if( __fwrite( stream, (void *)ptr, size*n, &writeCount ) == CE_GOOD )
		return writeCount / size; //return number of items read

	else
		return 0;
}

/******************************************************************************
 * Function:        int feof( FILE * stream )
 *
 * PreCondition:    File is open is read mode
 *
 * Input:           stream			- FILE * of the target file
 *
 * Output:          NON Zero		- EOF reached
 *                  Zero			- Not at end of File
 *
 * Side Effects:    None
 *
 * Overview:
 *
 * Note:            None
 *****************************************************************************/
int feof( FILE * stream )
{
	return( stream->seek == stream->size );
}

/******************************************************************************
 * Function:        size_t fread(void *ptr, size_t size, size_t n, FILE *stream)
 *
 * PreCondition:    FILEopen in the read mode has been executed
 *
 * Input:           stream		- File to be read
 *					ptr			- Destination buffer for read bytes
 *					n			- Number of units to be read
 *					size		- Size of units in bytes
 *
 * Output:			size_t		- number of units read
 *
 *
 * Side Effects:    None
 *
 * Overview:        Reads data from a file into destination buffer
 *
 * Note:            None
 *****************************************************************************/
size_t fread(void *ptr, size_t size, size_t n, FILE *stream)
{
	WORD readCount = 0;

	__fread( stream, ptr, size*n, &readCount );

	return readCount / size; //return number of items read
}


/******************************************************************************
 * Function:        int fclose(FILE *fo)
 *
 * PreCondition:    File opened
 *
 * Input:           fo		- Pointer to file structure
 *
 * Output:			0		- File closed successfully
 *					EOF		- File not closed successfully
 *
 * Side Effects:    None
 *
 * Overview:        Close a file
 *
 * Note:
 *****************************************************************************/
int fclose(FILE *fo)
{
	if( __fclose( fo ) == CE_GOOD )
		return 0;
	else
		return EOF;
}

/******************************************************************************
 * Function:        BYTE FormatFileName( const char* fileName, char* fN2 )
 *
 * PreCondition:    
 *
 * Input:           fileName	- The name to be formatted
 *					fN2			- The location the formatted name will be stored
 *
 * Output:			TRUE		- Name formatted successfully
 *					FALSE		- File name could not be formatted
 *
 * Side Effects:    None
 *
 * Overview:        Format an 8.3 filename into FILE structure format
 *
 * Note:
 *****************************************************************************/
BYTE FormatFileName( const char* fileName, char* fN2 )
{
	char * pExt;
	WORD temp;
//	WORD extLen;
	char szName[15];
	BYTE count;

	for (count = 0; count < 11; count++)
	{
		*(fN2 + count) = ' ';
	}

	temp = strlen( fileName );

	if( temp <= 12 ) //8+3+1
		strcpy( szName, fileName );  //copy to RAM in case fileName is located in flash
	else
		return FALSE; //long file name

	if ( !ValidateChars(szName) )
		return FALSE;

	//Look for '.' in the szName
	if( (pExt = strchr( szName, '.' )) != NULL )
	{
		*pExt = NULL;
		pExt++; // now pointing to extension

		if( strlen( pExt ) > 3 ) // make sure the extension is 3 bytes or less
			return FALSE;
	}

	if( strlen(szName) > 8 )
		return FALSE;

	//copy file name
	for (count = 0; count < strlen(szName); count++)
	{
		*(fN2 + count) = * (szName + count);
	}
	
	//copy extension
	if( pExt )
	{
		for (count = 0; count < strlen (pExt); count++)
		{
			*(fN2 + count + 8) = *(pExt + count);
		}
	}
	
	return TRUE;
}


/******************************************************************************
 * Function:        BYTE ValidateChars( char * FileName )
 *
 * PreCondition:    
 *
 * Input:           fileName	- The name to be validated
 *
 * Output:			TRUE		- Name was validated
 *					FALSE		- File name was not valid
 *
 * Side Effects:    None
 *
 * Overview:        Make sure a file name contains valid characters.	
 *
 * Note: 			Because of the way certain operating systems store
 *					file names, any file entry that is not in Long File
 *					name format can only contain upper case letters.
 *					If the specified name contains any lower case letters,
 *					this function will convert them to upper case
 *****************************************************************************/
BYTE ValidateChars( char * FileName )
{
	int StrSz, index;

	StrSz = strlen(FileName);

	for( index = 0; index < StrSz; index++ )
	{
		switch( isShort( FileName[index] ) )
		{

		default:
		case 0: // good char
			break;

		case 1: // lower case
			FileName[index] = toupper( FileName[index] );
			break;

		case 2: // invalid char or space.
		case 3:
			return FALSE;
			//FileName[index] = '_';
			//break;

		}
	}

	return TRUE;
}



/******************************************************************************
 * Function:        BYTE isShort (char aChar)
 *
 * PreCondition:    None
 *
 * Input:           aChar	- The char to be checked
 *                  
 * Output:			0		- The character is acceptable
 *					1		- The char is a lower case letter
 *					2		- Unacceptable char
 *                  3       - A space character
 *
 * Side Effects:    None
 *
 * Overview:        Checks to see if a char is acceptable in short format
 *
 * Note:            None
 *****************************************************************************/
BYTE isShort ( char aChar)
{
	if ((aChar >= 'A' && aChar <= 'Z') ||
			(aChar >= '0' && aChar <= '9') ||
			aChar == '$' || aChar == '%' || aChar == '\'' ||
			aChar == '-' || aChar == '_' || aChar == '@' ||
			aChar == '~' || aChar == '`' || aChar == '!' ||
			aChar == '(' || aChar == ')' || aChar == '^' ||
			aChar == '#' || aChar == '&' || aChar == '.')
		return 0;
	else
	{
		if (aChar >= 'a' && aChar <= 'z')
			return 1;
		else
		{
			if (aChar == ' ')
				return 3;
			else
				return 2;
		}
	}
}

/******************************************************************************
 * Function:        FILE * fopenpgm(const rom char * fileName, const rom char *mode)
 *
 * PreCondition:    None
 *
 * Input:           fileName	- The name of the file to be opened (ROM)
 *					mode		- The mode the file will be opened in (ROM)
 *                  
 * Output:			FILE *		- A pointer to the file object
 *
 * Side Effects:    None
 *
 * Overview:        Opens a file if the name and the mode are given in ROM 
 *					format.  Eg. fopenpgm ("FILE.TXT", "w");
 *
 * Note:            None
 *****************************************************************************/

#ifdef USE_PIC18
FILE * fopenpgm(const rom char * fileName, const rom char *mode)
#else
FILE * fopenpgm(const char * fileName, const char *mode)
#endif
{
	#ifdef USE_PIC18
		char F[13];
		char M[2];
		BYTE count;
			
	    *fileName;     // Load table pointer
	    for(count = 0; count < sizeof(F); count++)
	    {
	        _asm TBLRDPOSTINC _endasm
	        F[count] = TABLAT;
	    }//end for(...)
	
	    *mode;     // Load table pointer
	    for(count = 0; count < sizeof(M); count++)
	    {
	        _asm TBLRDPOSTINC _endasm
	        M[count] = TABLAT;
	    }//end for(...)
	
		return fopen(F, M);
	#else
		return fopen (fileName, mode);
	#endif
}

/******************************************************************************
 * Function:        int removepgm (const rom char * fileName)
 *
 * PreCondition:    None
 *
 * Input:           fileName	- The name of the file to be deleted (ROM)
 *                  
 * Output:			int			- 0 is successful, -1 otherwise
 *
 * Side Effects:    None
 *
 * Overview:        Removes a file if the name is given in ROM
 *					Eg. removepgm ("FILE.TXT");
 *
 * Note:            None
 *****************************************************************************/
#ifdef USE_PIC18
int removepgm (const rom char * fileName)
#else
int removepgm (const char * fileName)
#endif
{
	#ifdef USE_PIC18
		char F[13];
		BYTE count;
		
		*fileName;
		for(count = 0; count < sizeof(F); count++)
	    {
	        _asm TBLRDPOSTINC _endasm
	        F[count] = TABLAT;
	    }//end for(...)	
	
		return remove (F);
	#else
		return remove (fileName);
	#endif
}


WORD FATread (DISK *dsk, WORD ccls)
{
	BYTE i;
	DWORD p, l, li;
	WORD c;	

	if (dsk->type != FAT16)
		return CLUSTER_FAIL;

	p = (DWORD)ccls *2;
	
	l = dsk->fat + (p >> 9);

	p &= 0x1FF;

	if (lastFATSectorRead == l)
	{
		c = RAMreadW (FATBuffer, p);
	}
	else
	{
		// If there's a currently open FAT sector,
		// write it back before reading into the buffer
		if (needWrite)
		{
			for (i = 0, li = lastFATSectorRead; i < dsk->fatcopy; i++, li += dsk->fatsize)
				if (!SECTORwrite (lastFATSectorRead, FATBuffer))
					return FAIL;
			needWrite = FALSE;
		}

		if (!SECTORread (l, FATBuffer))
		{
			lastFATSectorRead = 0xFFFF;
			return CLUSTER_FAIL;
		}
		else
		{
			lastFATSectorRead = l;
			c = RAMreadW (FATBuffer, p);
		}
	}
	
	// Normalize it so 0xFFFF is an error
	if (c >= LAST_CLUSTER_FAT16)
		c = LAST_CLUSTER;
	
	return c;
}	// FATread


WORD FATwrite (DISK *dsk, WORD ccls, WORD value, BYTE forceWrite)
{
	BYTE i;
	DWORD p, li, l;

	if (dsk->type != FAT16)
		return FAIL;

	// The only purpose for calling this function with forceWrite
	// is to write the current FAT sector to the card
	if (forceWrite)
	{
		for (i = 0, li = lastFATSectorRead; i < dsk->fatcopy; i++, li += dsk->fatsize)
			if (!SECTORwrite (lastFATSectorRead, FATBuffer))
				return FAIL;

		needWrite = FALSE;

		return 0;
	}

	p = (DWORD) ccls *2;

	l = dsk->fat + (p >> 9);
	
	p &= 0x1FF;

	if (lastFATSectorRead != l)
	{
		// If we are loading a new sector then write
		// the current one to the card if we need to
		if (needWrite)
		{
			for (i = 0, li = lastFATSectorRead; i < dsk->fatcopy; i++, li += dsk->fatsize)
				if (!SECTORwrite (lastFATSectorRead, FATBuffer))
					return FAIL;
	
			needWrite = FALSE;
		}

		// Load the new sector
		if (!SECTORread (l, FATBuffer))
		{
			lastFATSectorRead = 0xFFFF;
			return FAIL;
		}
		else
			lastFATSectorRead = l;
	}

	RAMwrite (FATBuffer, p, value);				//lsb
	RAMwrite (FATBuffer, p+1, (value >> 8));	//msb

	needWrite = TRUE;

	return 0;
}
